/* game/match.c */
/* Matching routines for finding objects in room, inventory, or database. */

#include "externs.h"

int make_quiet;    // 1=Unmatched will not display "I don't see '...' here."
int match_count;   // Count multiple matches for the same abbreviation.
int match_accept;  // Looks for Nth item matching the abbreviation.


int match_thing(dbref player, char *name, int type)
{
  char shortname[512], *r, *s;
  dbref match_abs=NOTHING, exact_match=NOTHING, alt_location;
  int full, count=0, *ptr;

  if(Invalid(player))
    return NOTHING;
  if(!*name) {
    if(!(type & MATCH_QUIET) && !make_quiet)
      notify(player, "No object specified.");
    return NOTHING;
  }

  strmaxcpy(shortname, name, 512);

  /* set static variables. must we check for a player's full name? */
  match_count=0;
  full=(type & MATCH_FULLNAME)?1:0;

  /* find number to match */
  match_accept=1;
  if((s=strrchr(shortname, ' '))) {
    if(isnumber(s+1)) {
      match_accept=atoi(s+1);
      *s='\0';
    }
  }

  /* find absolute global database reference number */
  if(*name == '#' && isdigit(*(name+1)) && !Invalid(atoi(name+1)))
    match_abs = atoi(name+1);

  /* match dbref */
  if(type & MATCH_ABS)
    exact_match = match_abs;

  /* match an object in the zone of <player> */
  if((type & MATCH_INZONE) && exact_match == NOTHING && match_abs != NOTHING &&
     Typeof(player)==TYPE_ZONE && ((db[player].flags & ZONE_UNIVERSAL) ||
     inlist(db[mainroom(match_abs)].zone, player)))
    exact_match = match_abs;

  /* match player name */
  if((type & MATCH_PLAYER) && exact_match == NOTHING && *name == '*' &&
     (count=lookup_player(name+1)) > NOTHING)
    exact_match = count;

  /* match reflexive */
  if((type & MATCH_ME) && exact_match == NOTHING)
    if(!strcasecmp(name, "me"))
      exact_match = player;

  /* match home */
  if((type & MATCH_HOME) && exact_match == NOTHING)
    if(!strcasecmp(name, "home"))
      exact_match = HOME;

  /* match variable */
  if((type & MATCH_VARIABLE) && exact_match == NOTHING)
    if(!strcasecmp(name, "variable"))
      exact_match = AMBIGUOUS;

  /* match linked zone#s */
  if((type & MATCH_ZONE) && exact_match == NOTHING)
    for(ptr=db[mainroom(player)].zone;ptr && *ptr != NOTHING;ptr++)
      if(match_abs == *ptr)
        exact_match = *ptr;

  /* match container */
  if((type & MATCH_HERE) && exact_match == NOTHING)
    if(!strcasecmp(name, "here") && !Invalid(db[player].location))
      exact_match = db[player].location;

  /* match exits in room */
  if((type & MATCH_EXIT) && exact_match == NOTHING &&
     !Invalid(db[player].location))
    exact_match = match_exit(player, name, shortname, match_abs, type);

  /* match inventory objects */
  if((type & MATCH_INVEN) && exact_match == NOTHING)
    exact_match = match_list(player, player, name, shortname, match_abs,
                             type, &full);

  /* match contents in room */
  if((type & MATCH_CON) && exact_match == NOTHING &&
     !Invalid(db[player].location) && db[player].location != player) {
    count = match_list(player, db[player].location, name, shortname, match_abs,
                       type, &full);
    if(!Invalid(count))
      exact_match=count;
  }

  /* check possessive */
  if((type & MATCH_POSSESSIVE) && exact_match == NOTHING &&
     (s=strchr(shortname, '\'')) && s > shortname && (*(s+1)|32) == 's' &&
     *(s+2) == ' ') {
    int temp=match_accept;
    char buf[512];

    strcpy(buf, shortname);
    r=s-shortname+buf+3;
    *(r-3)='\0';

    while(*r == ' ')
      r++;

    /* match the contents of the possessive */
    if(*r && (alt_location=match_thing(player, buf, MATCH_QUIET|
       (type & ~MATCH_POSSESSIVE))) != NOTHING) {
      match_count=0;
      match_accept=temp;
      count = match_list(player, alt_location, r, r, NOTHING, type, &full);
      if(!Invalid(count))
        exact_match=count;
    }
  }

  /* destroyed objects do not exist */
  if(!Invalid(exact_match) && !(type&MATCH_DESTROYED) && Typeof(exact_match)
     == TYPE_GARBAGE)
    exact_match=NOTHING;

  /* return results */
  if(exact_match == NOTHING) {
    /* full name of the player not specified? */
    if(full == 2) {
      if(!(type & MATCH_QUIET))
        notify(player, "You must type out the player's name in full.");
      return AMBIGUOUS;
    } else if(!(type & MATCH_QUIET) && !make_quiet)
      notify(player, "I don't see '%s' here.", name);
  }

  return exact_match;
}

/* utility for matching objects or exits in room */
static int can_match(dbref player, dbref thing, int invis)
{
  /* check multi-dimensional plane */
  if(!match_plane(player, thing))
    return 0;

  /* matching self is always allowed, POW_DARK matches everything else */
  if(player == thing || (controls(player, thing, POW_DARK) &&
     !IS(player, TYPE_PLAYER, PUPPET)))
    return 1;

  if((Typeof(thing) == TYPE_PLAYER && !(db[thing].flags & PLAYER_CONNECT)) ||
     !could_doit(player, thing, A_VLOCK) || (!invis &&
     Typeof(thing) == TYPE_EXIT && !can_see(player, thing)))
    return 0;

  return 1;
}

int match_exit(player, match_name, match_prefix, absolute, type)
dbref player, absolute;
char *match_name, *match_prefix;
int type;
{
  dbref exit, cmatch=NOTHING;

  /* search room's exitlist */
  DOLIST(exit, db[db[player].location].exits) {
    /* if exit# matches absolute, finish up */
    if(exit == absolute)
      return exit;
    if(!can_match(player, exit, (type & MATCH_INVIS)))
      continue;

    /* if exit name matches full name */
    if(!strcasecmp(db[exit].name, match_name))
      return exit;
    if(!strcasecmp(db[exit].name, match_prefix)) {
      if(++match_count == match_accept)
        cmatch=exit;
      continue;
    }

    /* check exit aliases */
    if(exit_alias(exit, A_ALIAS, match_name))
      return exit;
    if(exit_alias(exit, A_ALIAS, match_prefix)) {
      if(++match_count == match_accept)
        cmatch=exit;
      break;
    }
  }

  return cmatch;
}

int match_list(player, loc, match_name, match_prefix, absolute, type, full)
dbref player, loc, absolute;
char *match_name, *match_prefix;
int type, *full;
{
  dbref first, cmatch=NOTHING;

  if(IS(loc, TYPE_ROOM, ROOM_XZONE))
    return NOTHING;

  DOLIST(first, db[loc].contents) {
    if(first == absolute)
      return first;
    if(!can_match(player, first, (type & MATCH_INVIS)))
      continue;

    /* match entire object name */
    if(!strcasecmp(db[first].name, match_name))
      return first;

    /* check partial object name */
    if(string_match(db[first].name, match_prefix)) {
      if(*full && Typeof(first) == TYPE_PLAYER)
        *full=2;
      else if(++match_count == match_accept)
        cmatch=first;
      continue;
    }

    if(Typeof(first) != TYPE_THING)
      continue;

    /* check object aliases */
    if(exit_alias(first, A_ALIAS, match_name))
      return first;
    if(exit_alias(first, A_ALIAS, match_prefix)) {
      if(++match_count == match_accept)
        cmatch=first;
    }
  }

  return cmatch;
}

/* accepts only nonempty matches starting at the beginning of a word */
char *string_match(char *src, char *sub)
{
  if(!*sub)
    return 0;

  while(*src) {
    if(string_prefix(src, sub))
      return src;

    /* else scan to beginning of next word */
    while(*src && *src != ' ' && *src != '-')
      src++;
    while(*src && (*src == ' ' || *src == '-'))
      src++;
  }
  return 0;
}

// Checks if a string matches a list of semicolon-separated exit aliases.
// Spaces surrounding the alias are ignored. Returns 1 if a match exists.
//
int exit_alias(dbref thing, unsigned char attr, char *str)
{
  char *r=atr_get(thing, attr), *s=r, *t;
  int len=strlen(str);

  while(*s) {
    t=s=strchrnul(r, ';');

    /* remove surrounding spaces */
    while(*r == ' ')
      r++;
    while(t > r && *(t-1) == ' ')
      t--;

    /* test for a match */
    if(t-r == len && !strncasecmp(r, str, t-r))
      return 1;
    r=s+1;
  }
  return 0;
}

// Matches a word from a space-separated list.
// Returns the word position of the match starting at 1, or 0 if no match.
//
int match_word(char *list, char *word)
{
  char *r=list, *s=r;
  int count=1, len=strlen(word);

  while(*s) {
    s=strchrnul(r, ' ');
    if(s-r == len && !strncasecmp(r, word, s-r))
      return count;
    r=s+1;
    count++;
  }
  return 0;
}

/* Wildcard Matching Procedures */

// d=string list; command typed in
// s=pattern to check; contains the wildcard symbols
//  flags-
// 1-10: record match data into %0 through %9 (n-1)
// Boolean (16):   check mathematical boolean args ! > < on *s
// Ansi (32):      don't strip ansi escape codes
// Recursive (64): recursive call to wild_match

int wild_match(char *d, char *s, int flags)
{
  static char buff[8192];
  static int wpos=0;
  int inv=0, rslt=0;
  char *t;

  /* set initialization parameters first time through */
  if(!(flags & WILD_RECURSIVE)) {
    /* set arg# to start recording match data */
    wpos=RANGE((flags & 15), 1, 10)-1;

    /* remove any stray ansi escape codes */
    if(!(flags & WILD_ANSI))
      d=ansi_strip(d);

    /* check for boolean arguments ! > < */
    if(flags & WILD_BOOLEAN) {
      if(*s == '!') {
        s++;
        inv=1;
      }
      if(*s == '<' || *s == '>') {
        s++;
        if(isdigit(*s) || (*s == '-' && isdigit(*(s+1))))
          rslt=(*(s-1) == '<')?(atoi(s) > atoi(d)):(atoi(s) < atoi(d));
        else {
          rslt=strcasecmp(s, d);
          rslt=(*(s-1) == '<')?(rslt > 0):(rslt < 0);
        }
        return inv ? !rslt:rslt;
      }
    }
  }

  /* match characters one at a time until we reach a '*' by itself */
  while(1) {
    if(*s != '*')
      while(*d == '\e')  /* Skip escape codes */
        skip_ansi(d);

    /* skip over excess wildcards '*?' or '**' */
    if(*s == '*') {
      s++;
      if(*s != '*' && *s != '?')
        break;
      continue;
    }

    /* single character match; returns false if at end of data */
    if(*s == '?') {
      if(!*d)
        return inv;
      if((flags & WILD_ENV) && wpos < 10) {
        wptr[wpos][0]=*d;
        wptr[wpos++][1]='\0';
      }
    } else {
      /* escape character; force literal match of next character */
      if(*s == '\\')
        s++;

      /* literal character match; if matching end of data, return true */
      if(tolower(*s) != tolower(*d))
        return inv;
      if(!*d)
        return !inv;
    }
    s++; d++;
  }

  /* return true on trailing '*' */
  if(!*s) {
    if((flags & WILD_ENV) && wpos < 10) {
      /* skip leading spaces */
      if(!(flags & WILD_ANSI))
        while(isspace(*d))
          d++;
      strcpy(wptr[wpos++], d);
    }
    return !inv;
  }

  /* scan for possible matches */
  t=buff;
  while(*d) {
    /* Skip or record ANSI escape codes */
    while(*d == '\e') {
      if(flags & WILD_ENV)
        copy_ansi(t,d);
      else
        skip_ansi(d);
    }
    if(!*d)
      break;

    if(*s == '\\')
      s++;
    if(tolower(*d) == tolower(*s) && wild_match(d+1, s+1, WILD_RECURSIVE)) {
      if((flags & WILD_ENV) && wpos < 10) {
        *t='\0';
        while(t-buff && isspace(*(t-1)))  /* remove following spaces */
          *--t='\0';
        strcpy(wptr[wpos++], buff);  /* record this * element */
        wild_match(d+1, s+1, flags|WILD_RECURSIVE); /* now record the others */
      }
      return !inv;
    } else if((flags & WILD_ENV) &&              /* skip leading spaces below */
              ((flags & WILD_ANSI) || t-buff || !isspace(*d))) {
      *t++=*d;
    }
    d++;
  }

  return inv;
}
