/* prog/hash.c */
/* dynamic hash table management */

#include "externs.h"

#define HASH_SIZE 239	// Size of table array. Relatively prime.


// Initialize a hash table with a pointer to a structure array.
//
// The first element in each array index must contain a distinct (char *)
// string by which to initialize the primary key of the hash table.
// The last array index must contain NULL for this (char *) field.
//
// entry=Pointer to first index in structure array.
// size=Size of 1 array item in struct.
// 
hash *make_hashtab(void *entry, int size)
{
  hash *op, *ptr;
  int i, sizes[HASH_SIZE]={0};
  
  size /= sizeof(hash *);
  op=malloc(HASH_SIZE * sizeof(hash *));

  /* go through the array and count the size of each bucket */
  for(ptr=entry;ptr->ptr;ptr+=size)
    sizes[hash_name((char *)ptr->ptr)]++;

  /* allocate space to hold items for each bucket */
  for(i=0;i<HASH_SIZE;i++) {
    op[i].ptr=malloc(sizeof(hash *)*(sizes[i]+1));
    op[i].ptr[sizes[i]].ptr=NULL;
  }

  /* go through the array again, this time setting the index pointers */
  for(ptr=entry;ptr->ptr;ptr+=size) {
    i=hash_name((char *)ptr->ptr);
    op[i].ptr[--sizes[i]].ptr=ptr;
  }

  return op;
}

// Attempt to return a unique number between 0 and HASH_SIZE-1 for each primary
// key (char *) field in the structure. Entries are case insensitive.
//
int hash_name(char *name)
{
  unsigned short op=0;
  int j, shift=0;
  char *s;
  
  for(s=name;*s;s++) {
    j = tolower(*s);
    op ^= j<<(shift);
    op ^= j>>(16-shift);
    shift = (shift+11)%16;
  }

  return op % HASH_SIZE;
}

// Look up an entry in the hash table by name. Return a pointer to the index
// of the array.
//
void *lookup_hash(hash *tab, char *name)
{
  int i=hash_name(name), j;

  for(j=0;tab[i].ptr[j].ptr;j++)
    if(!strcasecmp((char *)tab[i].ptr[j].ptr->ptr, name))
      return tab[i].ptr[j].ptr;
  return NULL;
}


// Create an empty hash table.
//
hash *create_hash()
{
  hash *op;
  int i;

  op=malloc(HASH_SIZE * sizeof(hash *));

  for(i=0;i<HASH_SIZE;i++) {
    op[i].ptr=malloc(sizeof(hash *));
    op[i].ptr[0].ptr=NULL;
  }

  return op;
}

// Add a pointer to structure to an existing hash table.
// Structure must not be part of an array.
//
void add_hash(hash *tab, void *entry)
{
  hash *ptr=entry;
  int j, i=hash_name((char *)ptr->ptr);

  /* count size of hash pointer */
  for(j=0;tab[i].ptr[j].ptr;j++);

  tab[i].ptr=realloc(tab[i].ptr, (j+2) * sizeof(hash *));
  tab[i].ptr[j].ptr=entry;
  tab[i].ptr[j+1].ptr=NULL;
}

// Delete an entry from a hash table, by name.
//
void del_hash(hash *tab, char *name)
{
  int j, loc=-1, i=hash_name(name);

  for(j=0;tab[i].ptr[j].ptr;j++)
    if(!strcasecmp((char *)tab[i].ptr[j].ptr->ptr, name))
      loc=j;
  if(loc == -1)
    return;

  /* free contents of structure (must not be part of an array) */
  free(tab[i].ptr[loc].ptr);

  if(j > 1)
    tab[i].ptr[loc].ptr=tab[i].ptr[j-1].ptr;
  tab[i].ptr[j-1].ptr=NULL;
  tab[i].ptr=realloc(tab[i].ptr, j * sizeof(hash *));
}
